import typing
from abc import abstractmethod

import attr

from applitools.common import CoordinatesType, FloatingMatchSettings, Point, Region
from applitools.core import GetFloatingRegion, GetRegion
from applitools.selenium.capture import EyesWebDriverScreenshot

if typing.TYPE_CHECKING:
    from typing import List
    from applitools.common import FloatingBounds
    from applitools.common.utils.custom_types import AnyWebElement, AnyWebDriver
    from applitools.selenium.selenium_eyes import SeleniumEyes

__all__ = (
    "RegionBySelector",
    "RegionByElement",
    "FloatingRegionBySelector",
    "FloatingRegionByElement",
)


def _region_from_element(element, screenshot):
    location = element.location
    size = element.size
    if screenshot:
        # Element's coordinates are context relative, so we need to convert them first.
        adjusted_location = screenshot.location_in_screenshot(
            Point(location["x"], location["y"]), CoordinatesType.CONTEXT_RELATIVE
        )
    else:
        adjusted_location = Point(location["x"], location["y"])
    region = Region.from_(adjusted_location, size)
    return region


class GetSeleniumRegion(GetRegion):
    def get_regions(self, eyes, screenshot):
        # type: (SeleniumEyes, EyesWebDriverScreenshot) -> List[Region]
        element = self._element(eyes.driver)
        region = _region_from_element(element, screenshot)
        return [region]

    def get_elements(self, driver):
        # type: (AnyWebDriver) -> List[AnyWebElement]
        return [self._element(driver)]

    @abstractmethod
    def _element(self, driver):
        pass


class GetSeleniumFloatingRegion(GetFloatingRegion):
    def get_regions(self, eyes, screenshot):
        # type: (SeleniumEyes, EyesWebDriverScreenshot) ->  List[FloatingMatchSettings]
        element = self._element(eyes.driver)
        region = _region_from_element(element, screenshot)
        return [FloatingMatchSettings(region, self.bounds)]

    def get_elements(self, driver):
        # type: (AnyWebDriver) -> List[AnyWebElement]
        return [self._element(driver)]

    @abstractmethod
    def _element(self, driver):
        pass


@attr.s
class RegionByElement(GetSeleniumRegion):
    element = attr.ib()  # type: AnyWebElement

    def _element(self, driver):
        return self.element


@attr.s
class RegionBySelector(GetSeleniumRegion):
    """
    :param by: The "by" part of a selenium selector for an element which
               represents the ignore region
    :param value: The "value" part of a selenium selector for
                  an element which represents the ignore region.
    """

    by = attr.ib()
    value = attr.ib()

    def _element(self, driver):
        return driver.find_element(self.by, self.value)


@attr.s
class FloatingRegionByElement(GetSeleniumFloatingRegion):
    """
    :ivar element: The element which represents the inner region (the floating part).
    :ivar bounds: The outer rectangle bounding the inner region.
    """

    element = attr.ib()  # type: AnyWebElement
    bounds = attr.ib()  # type: FloatingBounds

    def _element(self, driver):
        return self.element


@attr.s
class FloatingRegionBySelector(GetSeleniumFloatingRegion):
    """
    :ivar by: The selenium By
    :ivar value: The css selector for an element which represents the inner region.
    :ivar bounds: The outer rectangle bounding the inner region.
    """

    by = attr.ib()  # type: str
    value = attr.ib()  # type: str
    bounds = attr.ib()  # type: FloatingBounds

    def _element(self, driver):
        return driver.find_element(self.by, self.value)



import math
import typing

import attr
from PIL import Image
from selenium.common.exceptions import WebDriverException

from applitools.common import CoordinatesType, Point, RectangleSize, Region, logger
from applitools.common.utils import argument_guard, datetime_utils, image_utils
from applitools.core import PositionMemento, PositionProvider
from applitools.core.cut import NullCutProvider
from applitools.selenium.capture import EyesWebDriverScreenshot

if typing.TYPE_CHECKING:
    from typing import Optional, List, Dict
    from applitools.common import ScaleProvider
    from applitools.core.debug import DebugScreenshotProvider
    from applitools.common.geometry import SubregionForStitching
    from applitools.selenium.region_compensation import RegionPositionCompensation
    from applitools.core.capture import EyesScreenshotFactory, ImageProvider
    from applitools.core.cut import CutProvider


@attr.s
class FullPageCaptureAlgorithm(object):
    MIN_SCREENSHOT_PART_SIZE = 10

    wait_before_screenshots = attr.ib()  # type: int
    debug_screenshot_provider = attr.ib()  # type: DebugScreenshotProvider
    screenshot_factory = attr.ib()  # type: EyesScreenshotFactory
    origin_provider = attr.ib()  # type: PositionProvider
    scale_provider = attr.ib()  # type: ScaleProvider
    cut_provider = attr.ib()  # type: CutProvider
    stitch_overlap = attr.ib()  # type: int
    image_provider = attr.ib()  # type: ImageProvider
    region_position_compensation = attr.ib()  # type: RegionPositionCompensation

    def _debug_msg(self, msg):
        return "{}_{}".format(
            self.region_position_compensation._useragent.browser.name, msg
        )

    def get_stitched_region(self, region, full_area, position_provider):
        # type: (Region, Optional[Region], Optional[PositionProvider]) -> Image.Image
        argument_guard.not_none(region)
        argument_guard.not_none(position_provider)

        logger.info(
            "region: %s ; full_area: %s ; position_provider: %s"
            % (region, full_area, position_provider.__class__.__name__)
        )

        origin_state = self.origin_provider.get_state()

        if self.origin_provider != position_provider:
            self.origin_provider.set_position(
                Point.ZERO()
            )  # first scroll to 0,0 so CSS stitching works.

        # Saving the original position (in case we were already in the outermost frame).
        original_stitched_state = position_provider.get_state()

        datetime_utils.sleep(self.wait_before_screenshots)
        initial_screenshot = self.image_provider.get_image()
        initial_size = RectangleSize.from_(initial_screenshot)

        pixel_ratio = self._get_pixel_ratio(initial_screenshot)
        scaled_cut_provider = self.cut_provider.scale(pixel_ratio)
        cutted_initial_screenshot = self._cut_if_needed(
            initial_screenshot, scaled_cut_provider
        )
        self.debug_screenshot_provider.save(
            cutted_initial_screenshot, self._debug_msg("cutted_initial_screenshot")
        )

        region_in_initial_screenshot = self._get_region_in_screenshot(
            region, cutted_initial_screenshot, pixel_ratio
        )
        cropped_initial_screenshot = self._crop_if_needed(
            cutted_initial_screenshot, region_in_initial_screenshot
        )
        self.debug_screenshot_provider.save(
            cropped_initial_screenshot, self._debug_msg("cropped_initial_screenshot")
        )

        scaled_initial_screenshot = image_utils.scale_image(
            cropped_initial_screenshot, self.scale_provider
        )
        self.debug_screenshot_provider.save(
            scaled_initial_screenshot, self._debug_msg("scaled_initial_screenshot")
        )
        if full_area is None or full_area.is_empty:
            entire_size = self._get_entire_size(initial_screenshot, position_provider)
            # Notice that this might still happen even if we used
            # "get_image_part", since "entire_page_size" might be that of a
            # frame
            if (
                scaled_initial_screenshot.width >= entire_size.width
                and scaled_initial_screenshot.height >= entire_size.height
            ):
                self.origin_provider.restore_state(origin_state)
                return scaled_initial_screenshot

            full_area = Region.from_(Point.ZERO(), entire_size)

        scaled_cropped_location = full_area.location
        physical_crop_location = Point.from_(scaled_cropped_location).scale(pixel_ratio)

        if region_in_initial_screenshot.is_empty:
            physical_crop_size = RectangleSize(
                initial_size.width - physical_crop_location.x,
                initial_size.height - physical_crop_location.y,
            )
            source_region = Region.from_(physical_crop_location, physical_crop_size)
        else:
            # Starting with the screenshot we already captured at (0,0).
            source_region = region_in_initial_screenshot

        scaled_cropped_source_rect = self.cut_provider.to_region(source_region.size)
        scaled_cropped_source_rect = scaled_cropped_source_rect.offset(
            source_region.left, source_region.top
        )
        scaled_cropped_source_region = dict(
            x=int(math.ceil(scaled_cropped_source_rect.left / pixel_ratio)),
            y=int(math.ceil(scaled_cropped_source_rect.top / pixel_ratio)),
            width=int(math.ceil(scaled_cropped_source_rect.width / pixel_ratio)),
            height=int(math.ceil(scaled_cropped_source_rect.height / pixel_ratio)),
        )
        scaled_cropped_size = dict(
            width=scaled_cropped_source_region["width"],
            height=scaled_cropped_source_region["height"],
        )

        # Getting the list of viewport regions composing the page
        # (we'll take screenshot for each one).
        if region_in_initial_screenshot.is_empty:
            x = max(0, full_area.left)
            y = max(0, full_area.top)
            w = min(full_area.width, scaled_cropped_size["width"])
            h = min(full_area.height, scaled_cropped_size["height"])
            rect_in_initial_screenshot = Region(
                round(x * pixel_ratio),
                round(y * pixel_ratio),
                round(w * pixel_ratio),
                round(h * pixel_ratio),
            )
        else:
            rect_in_initial_screenshot = region_in_initial_screenshot

        screenshot_parts = self._get_image_parts(
            full_area, scaled_cropped_size, pixel_ratio, rect_in_initial_screenshot
        )

        # Starting with element region size part of the screenshot. Use it as a size
        # template.
        stitched_image = Image.new("RGBA", (full_area.width, full_area.height))
        # Take screenshot and stitch for each screenshot part.
        stitched_image = self._stitch_screenshot(
            original_stitched_state,
            position_provider,
            screenshot_parts,
            stitched_image,
            self.scale_provider.scale_ratio,
            scaled_cut_provider,
        )
        position_provider.restore_state(original_stitched_state)
        self.origin_provider.restore_state(origin_state)
        return stitched_image

    def _stitch_screenshot(
        self,
        original_stitch_state,  # type: PositionMemento
        stitch_provider,  # type: PositionProvider
        screenshot_parts,  # type: List[SubregionForStitching]
        stitched_image,  # type: Image.Image
        scale_ratio,  # type: float
        scaled_cut_provider,  # type: CutProvider
    ):
        # type: (...) -> Image
        logger.info("enter: scale_ratio {}".format(scale_ratio))
        for part_region in screenshot_parts:
            logger.debug("Part: {}".format(part_region))
            # Scroll to the part's top/left
            part_region_location = part_region.scroll_to.offset(
                original_stitch_state.position
            )
            origin_position = stitch_provider.set_position(part_region_location)

            target_position = part_region.paste_physical_location.offset(
                part_region_location - origin_position
            )

            # Actually taking the screenshot.
            datetime_utils.sleep(self.wait_before_screenshots)
            part_image = self.image_provider.get_image()
            self.debug_screenshot_provider.save(
                part_image, self._debug_msg("part_image")
            )

            cut_part = scaled_cut_provider.cut(part_image)
            self.debug_screenshot_provider.save(cut_part, self._debug_msg("cut_part"))

            r = part_region.physical_crop_area
            if not r.is_size_empty:
                cropped_part = image_utils.crop_image(cut_part, r)
            else:
                cropped_part = cut_part
            self.debug_screenshot_provider.save(
                cropped_part, self._debug_msg("cropped_part")
            )
            scaled_part_image = image_utils.scale_image(cropped_part, scale_ratio)
            self.debug_screenshot_provider.save(
                scaled_part_image, self._debug_msg("scaled_part_image")
            )

            r2 = part_region.logical_crop_area
            scaled_cropped_part_image = image_utils.crop_image(scaled_part_image, r2)
            self.debug_screenshot_provider.save(
                scaled_cropped_part_image, self._debug_msg("scaled_cropped_part_image")
            )

            logger.debug("pasting part at {}".format(target_position))
            image_utils.paste_image(
                stitched_image, scaled_cropped_part_image, target_position
            )
            self.debug_screenshot_provider.save(
                stitched_image, self._debug_msg("stitched_image")
            )
        return stitched_image

    def _crop_if_smaller(
        self,
        full_area,
        last_successful_location,
        last_successful_part_size,
        stitched_image,
    ):
        # type: (Region, Point, RectangleSize, Image) -> Image
        act_image_width = last_successful_location.x + last_successful_part_size.width
        act_image_height = last_successful_location.y + last_successful_part_size.height
        logger.info("Extracted entire size: {}".format(full_area.size))
        logger.info(
            "Actual stitched size: {} x {}".format(act_image_width, act_image_height)
        )
        self.debug_screenshot_provider.save(
            stitched_image, self._debug_msg("_stitched_before_trim")
        )
        if (
            act_image_width < stitched_image.width
            or act_image_height < stitched_image.height
        ):
            logger.info("Trimming unnecessary margins...")
            stitched_image = image_utils.get_image_part(
                stitched_image,
                Region(
                    0,
                    0,
                    min([act_image_width, stitched_image.width]),
                    min([act_image_height, stitched_image.height]),
                ),
            )
            logger.info("Done")
        return stitched_image

    def _get_entire_size(self, image, position_provider):
        # type: (Image, PositionProvider) -> RectangleSize
        try:
            entire_size = position_provider.get_entire_size()
            logger.info("Entire size of region context: {}".format(entire_size))
        except WebDriverException as e:
            logger.warning(
                "Failed to extract entire size of region context {}".format(e)
            )
            logger.debug(
                "Using image size instead: {} x()".format(image.width, image.height)
            )
            entire_size = RectangleSize(image.width, image.height)
        return entire_size

    def _crop_if_needed(self, image, region_in_screenshot):
        # type: (Image, Region) -> Image
        if not region_in_screenshot.is_size_empty:
            image = image_utils.get_image_part(image, region_in_screenshot)
        return image

    def _cut_if_needed(self, image, scaled_cut_provider):
        # type: (Image, NullCutProvider) -> Image
        if not isinstance(scaled_cut_provider, NullCutProvider):
            image = scaled_cut_provider.cut(image)
        return image

    def _get_pixel_ratio(self, image):
        # type: (Image) -> float
        self.scale_provider.update_scale_ratio(image.width)
        return 1.0 / self.scale_provider.scale_ratio

    def _get_image_parts(
        self, full_area, scaled_cropped_size, pixel_ratio, rect_in__initial_screenshot
    ):
        # type: (Region, Dict, float, Region) -> List[SubregionForStitching]
        # The screenshot part is a bit smaller than the screenshot size,
        # in order to eliminate duplicate bottom scroll bars, as well as fixed
        # position footers.
        part_image_size = RectangleSize(
            max(scaled_cropped_size["width"], self.MIN_SCREENSHOT_PART_SIZE),
            max(scaled_cropped_size["height"], self.MIN_SCREENSHOT_PART_SIZE),
        )
        logger.info(
            "entire page region: %s, image part size: %s" % (full_area, part_image_size)
        )
        # Getting the list of sub-regions composing the whole region (we'll take
        # screenshot for each one).
        image_parts = full_area.get_sub_regions(
            part_image_size,
            self.stitch_overlap,
            pixel_ratio,
            rect_in__initial_screenshot,
        )
        return image_parts

    def _get_region_in_screenshot(self, region, image, pixel_ratio):
        # type: (Region, Image, float) -> Region
        if region.is_size_empty:
            return region
        logger.info("Creating screenshot object...")
        #  We need the screenshot to be able to convert the region to screenshot coordinates.
        screenshot = self.screenshot_factory.make_screenshot(
            image
        )  # type: EyesWebDriverScreenshot
        logger.info("Getting region in screenshot...")

        region_in_screenshot = screenshot.intersected_region(
            region, CoordinatesType.SCREENSHOT_AS_IS
        )
        logger.info("Region in screenshot: {}".format(region_in_screenshot))
        region_in_screenshot = region_in_screenshot.scale(pixel_ratio)
        logger.info("Scaled region: {}".format(region_in_screenshot))

        region_in_screenshot = self.region_position_compensation.compensate_region_position(
            region_in_screenshot, pixel_ratio
        )
        # Handling a specific case where the region is actually larger than
        # the screenshot (e.g., when body width/height are set to 100%, and
        # an internal div is set to value which is larger than the viewport).
        region_in_screenshot = region_in_screenshot.intersect(Region.from_(image))
        logger.info("Region after intersect: {}".format(region_in_screenshot))
        return region_in_screenshot



from __future__ import absolute_import

import math
import typing as tp

from selenium.common.exceptions import WebDriverException
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement

from applitools.common import logger
from applitools.common.geometry import CoordinatesType, Point, Region
from applitools.common.utils.general_utils import proxy_to, all_fields

from . import eyes_selenium_utils

if tp.TYPE_CHECKING:
    from typing import Optional, Text
    from .positioning import SeleniumPositionProvider
    from .webdriver import EyesWebDriver

_JS_GET_COMPUTED_STYLE_FORMATTED_STR = """
            var elem = arguments[0];
            var styleProp = '%s';
            if (window.getComputedStyle) {
                return window.getComputedStyle(elem, null)
                .getPropertyValue(styleProp);
            } else if (elem.currentStyle) {
                return elem.currentStyle[styleProp];
            } else {
                return null;
            }
    """
_JS_GET_SCROLL_LEFT = "return arguments[0].scrollLeft;"
_JS_GET_SCROLL_TOP = "return arguments[0].scrollTop;"
_JS_GET_SCROLL_WIDTH = "return arguments[0].scrollWidth;"
_JS_GET_SCROLL_HEIGHT = "return arguments[0].scrollHeight;"
_JS_GET_OVERFLOW = "return arguments[0].style.overflow;"
_JS_SET_OVERFLOW_FORMATTED_STR = "arguments[0].style.overflow = '%s'"
_JS_GET_CLIENT_WIDTH = "return arguments[0].clientWidth;"
_JS_GET_CLIENT_HEIGHT = "return arguments[0].clientHeight;"
_JS_SCROLL_TO_FORMATTED_STR = """
        arguments[0].scrollLeft = {:d};
        arguments[0].scrollTop = {:d};
"""
_JS_GET_SCROLL_POSITION = """
return arguments[0].scrollLeft + ';' + arguments[0].scrollTop;
"""
_JS_GET_BORDER_WIDTHS_ARR = """
        var retVal = retVal || [];
        if (window.getComputedStyle) {
        var computedStyle = window.getComputedStyle(elem, null);
        retVal.push(computedStyle.getPropertyValue('border-left-width'));
        retVal.push(computedStyle.getPropertyValue('border-top-width'));
        retVal.push(computedStyle.getPropertyValue('border-right-width'));
        retVal.push(computedStyle.getPropertyValue('border-bottom-width'));
        } else if (elem.currentStyle) {
        retVal.push(elem.currentStyle['border-left-width']);
        retVal.push(elem.currentStyle['border-top-width']);
        retVal.push(elem.currentStyle['border-right-width']);
        retVal.push(elem.currentStyle['border-bottom-width']);
        } else {
        retVal.push(0,0,0,0);
        }
"""
_JS_GET_BORDER_WIDTHS = _JS_GET_BORDER_WIDTHS_ARR + "return retVal;"
_JS_GET_SIZE_AND_BORDER_WIDTHS = """
        var elem = arguments[0];
        var retVal = [arguments[0].clientWidth, arguments[0].clientHeight];
        {}
        return retVal
""".format(
    _JS_GET_BORDER_WIDTHS_ARR
)


@proxy_to("_element", all_fields(WebElement))
class EyesWebElement(object):
    """
    A wrapper for selenium web element. This enables eyes to be notified
     about actions/events for this element.
    """

    def __init__(self, element, driver):
        # type: (WebElement, EyesWebDriver) -> None
        """
        Ctor.

        :param element: The element in the frame.
        :param eyes: The eyes sdk instance.
        :param driver: EyesWebDriver instance.
        """
        if isinstance(element, EyesWebElement):
            element = element._element
        self._element = element  # type: WebElement
        self._driver = driver  # type: EyesWebDriver

        # setting from outside
        self.position_provider = None  # type: Optional[SeleniumPositionProvider]
        self._original_overflow = None  # type: Optional[Text]

    @property
    def element(self):
        # type: () -> WebElement
        return self._element

    @property
    def driver(self):
        # type: () -> EyesWebDriver
        return self._driver

    @property
    def bounds(self):
        # type: () -> Region
        # noinspection PyUnresolvedReferences
        location = self.location
        left, top = location["x"], location["y"]
        width = height = 0  # Default

        # noinspection PyBroadException
        try:
            size = self._element.size
            width, height = size["width"], size["height"]
        except Exception:
            # Not implemented on all platforms.
            pass
        if left < 0:
            left, width = 0, max(0, width + left)
        if top < 0:
            top, height = 0, max(0, height + top)
        return Region(left, top, width, height, CoordinatesType.CONTEXT_RELATIVE)

    @property
    def location(self):
        return self._element.location

    @property
    def rect(self):
        return self._element.rect

    @property
    def is_attached_to_page(self):
        try:
            self.location
            return True
        except WebDriverException:
            return False

    @property
    def size(self):
        return self._element.size

    @property
    def scroll_location(self):
        # type: () -> Point
        pos = self.driver.execute_script(_JS_GET_SCROLL_POSITION, self.element)
        return eyes_selenium_utils.parse_location_string(pos)

    @property
    def id(self):
        return self._element.id

    def value_of_css_property(self, property_name):
        return self._element.value_of_css_property(property_name)

    def find_element(self, by=By.ID, value=None):
        """
        Returns a WebElement denoted by "By".

        :param by: By which option to search for (default is by ID).
        :param value: The value to search for.
        :return: WebElement denoted by "By".
        """
        # Get result from the original implementation of the underlying driver.
        result = self._original_methods["find_element"](by, value)
        # Wrap the element.
        if result:
            result = EyesWebElement(result, self._driver)
        return result

    def find_elements(self, by=By.ID, value=None):
        """
        Returns a list of web elements denoted by "By".

        :param by: By which option to search for (default is by ID).
        :param value: The value to search for.
        :return: List of web elements denoted by "By".
        """
        # Get result from the original implementation of the underlying driver.
        results = self._original_methods["find_elements"](by, value)
        # Wrap all returned elements.
        if results:
            updated_results = []
            for element in results:
                updated_results.append(EyesWebElement(element, self._driver))
            results = updated_results
        return results

    def click(self):
        """
        Clicks and element.
        """
        self._driver.eyes.add_mouse_trigger_by_element("click", self)
        self._element.click()

    def send_keys(self, *value):
        """
        Sends keys to a certain element.

        :param value: The value to type into the element.
        """
        text = u""
        for val in value:
            if isinstance(val, int):
                val = val.__str__()
            text += val.encode("utf-8").decode("utf-8")
        self._driver.eyes.add_text_trigger_by_element(self, text)
        self._element.send_keys(*value)

    def hide_scrollbars(self):
        # type: () -> tp.Text
        """
        Hides the scrollbars of the current element.

        :return: The previous value of the overflow property (could be None).
        """
        logger.debug("EyesWebElement.hide_scrollbars()")
        self._original_overflow = eyes_selenium_utils.hide_scrollbars(  # type: ignore
            self.driver, self.element
        )
        return self._original_overflow

    def return_to_original_overflow(self):
        return eyes_selenium_utils.return_to_original_overflow(
            self.driver, self.element, self._original_overflow
        )

    def get_computed_style(self, prop_style):
        script = _JS_GET_COMPUTED_STYLE_FORMATTED_STR % prop_style
        return self._driver.execute_script(script, self._element)

    def get_computed_style_int(self, prop_style):

        value = self.get_computed_style(prop_style)
        return int(round(float(value.replace("px", "").strip())))

    def get_scroll_left(self):
        return int(
            math.ceil(self._driver.execute_script(_JS_GET_SCROLL_LEFT, self._element))
        )

    def get_scroll_top(self):
        return math.ceil(self._driver.execute_script(_JS_GET_SCROLL_TOP, self._element))

    def get_scroll_width(self):
        return int(
            math.ceil(self._driver.execute_script(_JS_GET_SCROLL_WIDTH, self._element))
        )

    def get_scroll_height(self):
        return int(
            math.ceil(self._driver.execute_script(_JS_GET_SCROLL_HEIGHT, self._element))
        )

    def get_border_left_width(self):
        return self.get_computed_style_int("border-left-width")

    def get_border_right_width(self):
        return self.get_computed_style_int("border-right-width")

    def get_border_top_width(self):
        return self.get_computed_style_int("border-right-width")

    def get_border_bottom_width(self):
        return self.get_computed_style_int("border-right-width")

    def get_overflow(self):
        return self._driver.execute_script(_JS_GET_OVERFLOW, self._element)

    def get_client_width(self):
        return int(
            math.ceil(
                float(self._driver.execute_script(_JS_GET_CLIENT_WIDTH, self._element))
            )
        )

    def get_client_height(self):
        return int(
            math.ceil(
                float(self._driver.execute_script(_JS_GET_CLIENT_HEIGHT, self._element))
            )
        )

    def scroll_to(self, location):
        # type: (Point) -> Point
        """Scrolls to the specified location inside the element."""
        position = self._driver.execute_script(
            _JS_SCROLL_TO_FORMATTED_STR.format(location["x"], location["y"])
            + _JS_GET_SCROLL_POSITION,
            self._element,
        )
        return eyes_selenium_utils.parse_location_string(position)

    @property
    def size_and_borders(self):
        # type: () -> SizeAndBorders
        ret_val = self._driver.execute_script(
            _JS_GET_SIZE_AND_BORDER_WIDTHS, self._element
        )
        return SizeAndBorders(
            width=int(ret_val[0]),
            height=int(ret_val[1]),
            left=int(ret_val[2].rstrip("px")),
            top=int(ret_val[3].rstrip("px")),
            right=int(ret_val[4].rstrip("px")),
            bottom=int(ret_val[5].rstrip("px")),
        )

    @property
    def bounding_client_rect(self):
        left, top, width, height = self.driver.execute_script(
            "var r = arguments[0].getBoundingClientRect();"
            "return r.left+';'+r.top+';'+r.width+';'+r.height;",
            self._element,
        ).split(";")
        return dict(
            x=math.ceil(float(left)),
            y=math.ceil(float(top)),
            width=math.ceil(float(width)),
            height=math.ceil(float(height)),
        )

    def __str__(self):
        tag_name = "None"
        if self.is_attached_to_page:
            tag_name = self._element.tag_name
        return "EyesWebElement: id {}, tag_name {}".format(self._element.id, tag_name)


class SizeAndBorders(object):
    __slots__ = ("size", "borders")

    def __init__(self, width, height, left, top, right, bottom):
        self.size = dict(width=width, height=height)
        self.borders = dict(left=left, top=top, right=right, bottom=bottom)

    def __str__(self):
        return "SizeAndBorders(size={}, borders={})".format(self.size, self.borders)



from __future__ import absolute_import

import typing

import attr
from selenium.common.exceptions import WebDriverException

from applitools.common import (
    CoordinatesType,
    CoordinatesTypeConversionError,
    EyesError,
    OutOfBoundsError,
    Point,
    RectangleSize,
    Region,
    logger,
)
from applitools.common.utils import argument_guard, image_utils
from applitools.core.capture import EyesScreenshot, EyesScreenshotFactory
from applitools.selenium import eyes_selenium_utils
from applitools.selenium.capture.screenshot_utils import (
    ScreenshotType,
    calc_frame_location_in_screenshot,
    update_screenshot_type,
)
from applitools.selenium.frames import FrameChain

if typing.TYPE_CHECKING:
    from typing import Union, Optional
    from PIL import Image
    from applitools.selenium.webdriver import EyesWebDriver
    from applitools.selenium.positioning import SeleniumPositionProvider


@attr.s
class EyesWebDriverScreenshot(EyesScreenshot):

    _driver = attr.ib()  # type: EyesWebDriver
    _image = attr.ib()  # type: Image.Image
    _screenshot_type = attr.ib()  # type: ScreenshotType
    _frame_location_in_screenshot = attr.ib()  # type: Point
    _current_frame_scroll_position = attr.ib(default=None)  # type: Optional[Point]
    frame_window = attr.ib(default=None)  # type: Region
    region_window = attr.ib(default=Region(0, 0, 0, 0))  # type: Region
    _frame_chain = attr.ib(init=False)  # type: FrameChain

    @classmethod
    def create_viewport(cls, driver, image):
        # type: (EyesWebDriver, Image.Image) -> EyesWebDriverScreenshot
        instance = cls(driver, image, ScreenshotType.VIEWPORT, None)
        instance._validate_frame_window()
        return instance

    @classmethod
    def create_full_page(cls, driver, image, frame_location_in_screenshot):
        # type: (EyesWebDriver, Image.Image, Point) -> EyesWebDriverScreenshot
        return cls(driver, image, None, frame_location_in_screenshot)

    @classmethod
    def create_entire_frame(cls, driver, image, entire_frame_size):
        # type: (EyesWebDriver, Image.Image, RectangleSize) -> EyesWebDriverScreenshot
        return cls(
            driver,
            image,
            ScreenshotType.ENTIRE_FRAME,
            frame_location_in_screenshot=Point(0, 0),
            current_frame_scroll_position=Point(0, 0),
            frame_window=Region.from_(Point(0, 0), entire_frame_size),
        )

    @classmethod
    def from_screenshot(cls, driver, image, screenshot_region):
        # type: (EyesWebDriver, Image.Image, Region) -> EyesWebDriverScreenshot
        return cls(
            driver,
            image,
            ScreenshotType.ENTIRE_FRAME,
            Point.ZERO(),
            frame_window=Region.from_(Point.ZERO(), screenshot_region.size),
            region_window=Region.from_(screenshot_region),
        )

    def __attrs_post_init__(self):
        # type: () -> None
        self._screenshot_type = update_screenshot_type(
            self._screenshot_type, self._image, self._driver
        )
        position_provider = eyes_selenium_utils.get_cur_position_provider(self._driver)

        if not self._driver.is_mobile_app:
            self._frame_chain = self._driver.frame_chain.clone()
            frame_size = self.get_frame_size(position_provider)
            self._current_frame_scroll_position = eyes_selenium_utils.get_updated_scroll_position(  # noqa
                position_provider
            )
            self.updated_frame_location_in_screenshot(
                self._frame_location_in_screenshot
            )
            logger.debug("Calculating frame window...")
            self.frame_window = Region.from_(
                self._frame_location_in_screenshot, frame_size
            )
        else:
            self._frame_chain = FrameChain()
            self._current_frame_scroll_position = Point.ZERO()
            self._frame_location_in_screenshot = Point.ZERO()
            self.frame_window = Region.from_(
                self._frame_location_in_screenshot, self.image
            )
        self.frame_window = self.frame_window.intersect(Region.from_(self.image))

    def _validate_frame_window(self):
        # type: () -> None
        if self.frame_window.width <= 0 or self.frame_window.height <= 0:
            raise EyesError("Got empty frame window for screenshot!")

    def updated_frame_location_in_screenshot(self, location):
        # type: (Point) -> None
        if location is None:
            if self.frame_chain.size > 0:
                self._frame_location_in_screenshot = calc_frame_location_in_screenshot(
                    self._driver, self._frame_chain, self._screenshot_type
                )
            else:
                self._frame_location_in_screenshot = Point.ZERO()
        else:
            self._frame_location_in_screenshot = location

    @property
    def image(self):
        # type: () -> Union[Image, Image]
        return self._image

    def get_frame_size(self, position_provider):
        # type: (SeleniumPositionProvider) -> RectangleSize
        if self._frame_chain:
            frame_size = self._frame_chain.peek.outer_size
        else:
            # get entire page size might throw an exception for applications
            # which don't support Javascript (e.g., Appium). In that case
            # we'll use the viewport size as the frame's size.
            try:
                frame_size = position_provider.get_entire_size()
            except WebDriverException:
                # For Appium, we can't get the "entire page size",
                # so we use the viewport size.
                frame_size = self._driver.get_default_content_viewport_size()
        return frame_size

    @property
    def frame_chain(self):
        # type: () -> FrameChain
        return self._frame_chain

    def location_in_screenshot(self, location, coordinates_type):
        location = self.convert_location(
            location, coordinates_type, self.SCREENSHOT_AS_IS
        )
        # Making sure it's within the screenshot bounds
        if not self.frame_window.contains(location):
            raise OutOfBoundsError(
                "Location %s ('%s') is not visible in screenshot!"
                % (location, coordinates_type)
            )
        return location

    def sub_screenshot(self, region, throw_if_clipped=False):
        # We calculate intersection based on as-is coordinates.
        as_is_sub_screenshot_region = self.intersected_region(
            region, self.SCREENSHOT_AS_IS
        )
        if as_is_sub_screenshot_region.is_size_empty or (
            throw_if_clipped and not as_is_sub_screenshot_region.size == region.size
        ):
            raise OutOfBoundsError(
                "Region [%s] is out of screenshot bounds [%s]"
                % (region, self.frame_window)
            )
        sub_image = image_utils.get_image_part(self.image, as_is_sub_screenshot_region)
        result = EyesWebDriverScreenshot.from_screenshot(
            self._driver,
            sub_image,
            Region(region.left, region.top, sub_image.width, sub_image.height),
        )
        return result

    CONTEXT_RELATIVE = CoordinatesType.CONTEXT_RELATIVE
    SCREENSHOT_AS_IS = CoordinatesType.SCREENSHOT_AS_IS
    CONTEXT_AS_IS = CoordinatesType.CONTEXT_AS_IS

    def convert_location(self, location, from_, to):  # noqa: C901
        # type: (Point, CoordinatesType, CoordinatesType) -> Point
        argument_guard.not_none(location)
        argument_guard.not_none(from_)
        argument_guard.not_none(to)

        result = Point.from_(location)
        if from_ == to:
            return result

        # If we're not inside a frame, and the screenshot is the entire
        # page, then the context as-is/relative are the same (notice
        # screenshot as-is might be different, e.g.,
        # if it is actually a sub-screenshot of a region).
        if (
            len(self.frame_chain) == 0
            and self._screenshot_type == ScreenshotType.ENTIRE_FRAME
        ):
            if (
                from_ == self.CONTEXT_RELATIVE or from_ == self.CONTEXT_AS_IS
            ) and to == self.SCREENSHOT_AS_IS:
                # If this is not a sub-screenshot, this will have no effect.
                result = result.offset(self._frame_location_in_screenshot)
                # If this is not a region subscreenshot, this will have no effect.
                result = result.offset(
                    -self.region_window.left, -self.region_window.top
                )
            elif from_ == self.SCREENSHOT_AS_IS and (
                to == self.CONTEXT_RELATIVE or to == self.CONTEXT_AS_IS
            ):
                result = result.offset(-self._frame_location_in_screenshot)
            return result

        if from_ == self.CONTEXT_AS_IS:
            if to == self.CONTEXT_RELATIVE:
                result = result.offset(self._current_frame_scroll_position)
            elif to == self.SCREENSHOT_AS_IS:
                result = result.offset(self._frame_location_in_screenshot)
            else:
                raise CoordinatesTypeConversionError(from_, to)
        elif from_ == self.CONTEXT_RELATIVE:
            if to == self.SCREENSHOT_AS_IS:
                # First, convert context-relative to context-as-is.
                result = result.offset(-self._current_frame_scroll_position)
                # Now convert context-as-is to screenshot-as-is
                result = result.offset(self._frame_location_in_screenshot)
            elif to == self.CONTEXT_AS_IS:
                result = result.offset(-self._current_frame_scroll_position)
            else:
                raise CoordinatesTypeConversionError(from_, to)
        elif from_ == self.SCREENSHOT_AS_IS:
            if to == self.CONTEXT_RELATIVE:
                # First, convert context-relative to context-as-is.
                result = result.offset(-self._frame_location_in_screenshot)
                # Now convert context-as-is to screenshot-as-is
                result = result.offset(self._current_frame_scroll_position)
            elif to == self.CONTEXT_AS_IS:
                result = result.offset(-self._frame_location_in_screenshot)
            else:
                raise CoordinatesTypeConversionError(from_, to)
        else:
            raise CoordinatesTypeConversionError(from_, to)
        return result

    def intersected_region(self, region, coordinates_type):
        # type: (Region, CoordinatesType) -> Region
        argument_guard.not_none(region)
        argument_guard.not_none(coordinates_type)

        if region.is_size_empty:
            return Region.from_(region)

        original_coordinates_type = region.coordinates_type

        intersected_region = self.convert_region_location(
            region, original_coordinates_type, self.SCREENSHOT_AS_IS
        )
        #  If the request was context based, we intersect with the frame window.
        if (
            original_coordinates_type == self.CONTEXT_AS_IS
            or original_coordinates_type == self.CONTEXT_RELATIVE
        ):
            intersected_region = intersected_region.intersect(self.frame_window)
        # If the request is screenshot based, we intersect with the image
        elif original_coordinates_type == self.SCREENSHOT_AS_IS:
            intersected_region = intersected_region.intersect(
                Region(0, 0, self.image.width, self.image.height)
            )
        else:
            raise ValueError(
                "Unknown coordinates type: '%s'" % original_coordinates_type
            )
        #  If the intersection is empty we don't want to convert the coordinates.
        if intersected_region.is_size_empty:
            return intersected_region

        #  Converting the result to the required coordinates type
        intersected_region = self.convert_region_location(
            intersected_region, self.SCREENSHOT_AS_IS, coordinates_type
        )
        return intersected_region


@attr.s
class EyesWebDriverScreenshotFactory(EyesScreenshotFactory):
    """
    Encapsulates the instantiation of an `EyesWebDriverScreenshot`
    """

    _driver = attr.ib()

    def make_screenshot(self, image):
        # type: (Image) -> EyesWebDriverScreenshot
        return EyesWebDriverScreenshot.create_viewport(self._driver, image)
